use crate::error::Error;
use std::{
    convert::TryFrom,
    fmt::{Display, Formatter},
};

use super::Relocatable;

#[derive(Clone)]
pub struct SectionInfo {
    pub object_index: usize,
    pub section_index: usize,
    pub file_size: u64,
    pub data_size: u64,
    pub offset: u64,
}

#[derive(Clone)]
pub enum Section {
    Text(SectionInfo),
    Data(SectionInfo, bool), // readonly bool
    Bss(SectionInfo),
}

impl Section {
    pub fn data<'o, R>(&self, object: &'o R) -> Result<&'o [u8], Error>
    where
        R: Relocatable,
    {
        object.bytes(self.offset(), self.file_size())
    }

    pub fn object_index(&self) -> usize {
        self.info().object_index
    }

    pub fn offset(&self) -> u64 {
        self.info().offset
    }

    pub fn data_size(&self) -> u64 {
        self.info().data_size
    }

    pub fn file_size(&self) -> u64 {
        self.info().file_size
    }

    fn info(&self) -> &SectionInfo {
        match self {
            Section::Text(si) => si,
            Section::Data(si, _) => si,
            Section::Bss(si) => si,
        }
    }
}

impl<'data> TryFrom<(&str, SectionInfo)> for Section {
    type Error = Error;

    fn try_from(value: (&str, SectionInfo)) -> Result<Self, Self::Error> {
        if value.0.starts_with(".text") {
            Ok(Section::Text(value.1))
        } else if value.0.starts_with(".rodata") {
            Ok(Section::Data(value.1, true))
        } else if value.0.starts_with(".data") {
            Ok(Section::Data(value.1, false))
        } else if value.0.starts_with(".bss") {
            Ok(Section::Bss(value.1))
        } else {
            Err(Error::InvalidSectionName)
        }
    }
}

impl Display for SectionInfo {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "file_size: {}, data_size: {}, data: {}, offset: {}",
            self.file_size,
            self.data_size,
            self.data_size > 0,
            self.offset,
        )
    }
}

impl Display for Section {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            Section::Text(s) => writeln!(f, "TEXT {}", s),
            Section::Data(s, false) => writeln!(f, "DATA {}", s),
            Section::Data(s, true) => writeln!(f, "RO-DATA {}", s),
            Section::Bss(s) => writeln!(f, "BSS {}", s),
        }
    }
}
