use std::fmt::Debug;

use crate::error::{Error, LinkError, Trace, LE_GLOBAL_SYMBOL_DUPLICATE};

use super::Relocatable;

#[derive(Default, Debug, Clone, Copy, PartialEq, Eq)]
pub struct SymbolIndex {
    pub object_index: usize,
    pub section_index: usize,
    pub symbol_index: usize,
    pub offsets: SymbolOffsets,
}

#[derive(Default, Debug, Clone, Copy, PartialEq, Eq)]
pub struct SymbolOffsets {
    pub section_offset: usize, // within the source relocatable
    pub symbol_offset: usize,  // within the section
}

impl SymbolOffsets {
    pub fn file_offset(&self) -> usize {
        self.section_offset + self.symbol_offset
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum SymbolValue<'o> {
    Undefined,
    Address(usize),
    Value(&'o [u8]),
}

#[derive(Default)]
pub struct Symbols {
    indexes: Vec<SymbolIndex>,
    value_index: Option<usize>,
}

impl Symbols {
    pub fn push<R>(&mut self, si: SymbolIndex, objects: &[R]) -> Result<(), Error>
    where
        R: Relocatable,
    {
        let object = objects
            .get(si.object_index)
            .ok_or(Error::InvalidSymbolIndex)?;
        let value = object.symbol_value(si.symbol_index)?;

        self.indexes.push(si);

        // save "value index" if we found a value
        match value {
            SymbolValue::Value(_) | SymbolValue::Address(_) => {
                if let Some(existing) = self.value_index {
                    let osi = self
                        .indexes
                        .get(existing)
                        .ok_or(Error::InvalidSymbolIndex)?;
                    Self::duplicate_symbol_error(si, *osi, objects)
                } else {
                    self.value_index = Some(self.indexes.len() - 1);

                    Ok(())
                }
            }
            SymbolValue::Undefined => Ok(()),
        }
    }

    pub fn value<'o, R>(&self, objects: &'o [R]) -> Result<SymbolValue<'o>, Error>
    where
        R: Relocatable,
    {
        self.value_index.map_or(Ok(SymbolValue::Undefined), |v| {
            let si = self.indexes[v];
            // can this be compile time checked??
            if let Some(object) = objects.get(si.object_index) {
                object.symbol_value(si.symbol_index)
            } else {
                Err(Error::InvalidObjectIndex)
            }
        })
    }

    pub fn unresolved(&self) -> impl Iterator<Item = &SymbolIndex> {
        let vi = self.value_index.unwrap_or(std::usize::MAX);

        self.indexes
            .iter()
            .enumerate()
            .filter_map(move |(i, s)| match i == vi {
                true => None,
                false => Some(s),
            })
    }

    fn duplicate_symbol_error<R>(
        si: SymbolIndex,
        osi: SymbolIndex,
        objects: &[R],
    ) -> Result<(), Error>
    where
        R: Relocatable,
    {
        let object = objects
            .get(si.object_index)
            .ok_or(Error::InvalidSymbolIndex)?;
        let other = objects
            .get(osi.object_index)
            .ok_or(Error::InvalidObjectIndex)?;
        let mut traces = Vec::new();

        traces.push(Trace {
            origin: object.origin().into(),
            offset: object.symbol_offsets(si.symbol_index)?.file_offset(),
            source_info: None,
        });

        traces.push(Trace {
            origin: other.origin().into(),
            offset: other.symbol_offsets(osi.symbol_index)?.file_offset(),
            source_info: None,
        });

        let link_error = LinkError {
            code: LE_GLOBAL_SYMBOL_DUPLICATE,
            message: "duplicate global symbol found".into(),
            traces,
        };

        Err(Error::LinkingError(link_error))
    }
}
