use std::{
    fmt::{Debug, Display, Formatter},
    path::PathBuf,
};

pub const LE_GLOBAL_SYMBOL_DUPLICATE: u32 = 10001;

#[derive(Debug)]
pub enum Error {
    IOError(Box<dyn std::error::Error>),
    DataError(PathBuf), // generic "something wrong with data" error, not specific to parsing
    InvalidObjectIndex,
    InvalidObjectType(u32),
    InvalidSectionIndex,
    InvalidSectionName,
    InvalidSymbolIndex,
    InvalidSymbolName,
    InvalidSectionData,
    InvalidInput, // input is completely wrong (e.g. no magic bytes etc.)
    ParseError(Box<dyn std::error::Error>), // error on parsing input
    TryFromIntError,
    LinkingError(LinkError),
}

#[derive(Debug, Default)]
pub struct Trace {
    pub origin: PathBuf,
    pub offset: usize, // 0 indicates unknown/invalid
    pub source_info: Option<Box<dyn SourceInfo>>,
}

// "backtrace" for error origin
#[derive(Debug, Default)]
pub struct LinkError {
    pub code: u32, // specific trace code, should be >= 10000
    pub message: String,
    pub traces: Vec<Trace>,
}

impl Error {
    // used for programmatic input/output handling, IDE's etc.
    pub fn code(&self) -> u32 {
        match self {
            Error::IOError(_) => 100,
            Error::DataError(_) => 800,
            Error::InvalidObjectIndex => 200,
            Error::InvalidObjectType(_) => 201,
            Error::InvalidSectionIndex => 300,
            Error::InvalidSectionName => 301,
            Error::InvalidSectionData => 302,
            Error::InvalidSymbolIndex => 400,
            Error::InvalidSymbolName => 401,
            Error::ParseError(_) => 500,
            Error::TryFromIntError => 600,
            Error::InvalidInput => 700,
            Error::LinkingError(le) => le.code,
        }
    }
}

pub trait SourceInfo: Debug + Display {
    fn line(&self) -> u64;

    // todo
}

impl From<std::io::Error> for Error {
    fn from(err: std::io::Error) -> Self {
        Self::IOError(Box::new(err))
    }
}

impl From<std::num::TryFromIntError> for Error {
    fn from(_: std::num::TryFromIntError) -> Self {
        Self::TryFromIntError
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "[{}] ", self.code())?;

        match self {
            Error::IOError(err) => write!(f, "IO error: {}", err),
            Error::DataError(origin) => write!(f, "Data error in {:?}", origin),
            Error::InvalidObjectIndex => write!(f, "Invalid section index"),
            Error::InvalidObjectType(ot) => write!(f, "Invalid object type: {}", ot),
            Error::InvalidSectionIndex => write!(f, "Invalid section index"),
            Error::InvalidSectionName => write!(f, "Invalid section name"),
            Error::InvalidSectionData => write!(f, "Invalid section data"),
            Error::InvalidSymbolIndex => write!(f, "Invalid symbol index"),
            Error::InvalidSymbolName => write!(f, "Invalid symbol name"),
            Error::ParseError(err) => write!(f, "Parse error: {}", err),
            Error::TryFromIntError => write!(f, "Integer conversion error"),
            Error::InvalidInput => write!(f, "Input object invalid"),
            Error::LinkingError(trace) => write!(f, "Linking error: {}", trace),
        }
    }
}

impl Display for LinkError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let mut new_line = ""; // new line in case of previous info only

        for trace in &self.traces {
            if let Some(origin) = trace.origin.to_str() {
                new_line = "\n";
                write!(f, "\n\t{}", origin)?;
            }

            if trace.offset > 0 {
                new_line = "\n";
                write!(f, "\n@offset: {}", trace.offset)?;
            }

            if let Some(si) = &trace.source_info {
                new_line = "\n";
                write!(f, "\n{}", si)?;
            }
        }

        write!(f, "{}{}", new_line, self.message)
    }
}
